// -------------------------------------------------------------------------------------------
// Basic Master
// -------------------------------------------------------------------------------------------
//
//
// -------------------------------------------------------------------------------------------

#include <i2c_t3.h>
#include <ADC.h>

// These constants won't change:
// Buttons
const int button1 = 24;
const int button2 = 27;
const int button3 = 30;
const int button4 = 32;
const int button5 = 11;
const int sensorPin = A0;    // pin that the sensor is attached to
const int sensorPin2 = A20;
const int ledPin_calib = 9;        // pin that the LED is attached to
const int ledPin_adc = 8;        // pin that the LED is attached to

// variables:
int sensorMin = 1023;        // minimum sensor value
int sensorMax = 0;           // maximum sensor value

// Memory
bool isRead;
bool isTempADC;
bool isPressing;
uint8_t currAddr;
uint16_t rawTwoByte;
uint8_t writeArray[2];
uint8_t target = 0x40; // target Slave address of PGA309
uint8_t byte1; uint8_t byte2;

const unsigned long bufferAmt = 20000;
int bufferPacket[bufferAmt];
long timePacket[bufferAmt];

ADC *adc = new ADC();

void setup()
{
  pinMode(sensorPin,INPUT);
  pinMode(LED_BUILTIN,OUTPUT);    // LED
  pinMode(ledPin_calib,OUTPUT);    // LED
  pinMode(ledPin_adc,OUTPUT);    // LED
  digitalWrite(LED_BUILTIN,LOW);  // LED off
  digitalWrite(ledPin_calib,LOW);    // LED off
  digitalWrite(ledPin_adc,LOW);    // LED off
  
  // internal resistor enable for 
  pinMode(button1,INPUT_PULLUP);       // Control for Send/Receive select
  pinMode(button2,INPUT_PULLUP);       // Control for address
  pinMode(button3,INPUT_PULLUP);       // Start address program
  pinMode(button4,INPUT_PULLUP);       // Send ADC readings
  pinMode(button5,INPUT_PULLUP);       // Calibrate

  //8,10,12,16
  adc->setResolution(16);
  //0,4,8,16,32
  adc->setAveraging(4);
  adc->setSamplingSpeed(ADC_MED_SPEED);
  adc->setConversionSpeed(ADC_MED_SPEED);
  
  // Setup for Master mode, pins 18/19, external pullups, 400kHz, 200ms default timeout
  Wire.begin(I2C_MASTER, target, I2C_PINS_18_19, I2C_PULLUP_EXT, 400000);
  Wire.setDefaultTimeout(200000); // 200ms

  isPressing = false;
  isRead = false;
  isTempADC = true;
  currAddr = 0x00;
  byte1 = 0x00;
  byte2 = 0x00;

  Serial.begin(115200);
  delay(100);
}


// Incorporate button interrupts?
void loop(){

    digitalWrite(LED_BUILTIN,LOW);  // LED on
    if(digitalRead(button1) == LOW){
      isPressing = true;
      isRead = !isRead;
      Serial.print("Read/Write state switched: \n");
      if(isRead)
      Serial.print("Set to Read\n");
      else
      Serial.print("Set to Write\n");
      Serial.print("\n");
      delay(800);
    }

    if(digitalRead(button2) == LOW){
      digitalWrite(ledPin_adc,HIGH);    // LED off
      Serial.println("ADC data being sent to MatLab.");
      sendADCData();
      Serial.println("ADC data transmission done.");
      digitalWrite(ledPin_adc,LOW);    // LED off
    }
}

// Send data through USB
void sendADCData(){
  //long i = 0; 
  unsigned long j = 0;
  int adcReading;

  unsigned long duration = 1000; // 1 sec
  unsigned long startTime = millis();
  while(millis() < (startTime + duration)){
    adcReading = adc->analogRead(sensorPin);
    bufferPacket[j] = adcReading;
    timePacket[j] = micros() - (startTime * 1000);
    j = j + 1;
    if (j == bufferAmt){
      Serial.println("Buffer has been filled");
      unsigned long elapsed = millis() - startTime;
      Serial.printf("The sampling time was %d ms. \n",elapsed);
      break;
    }
  }
  while(millis() < (startTime + 2*duration));
  for (unsigned long k=0;k<bufferAmt;k++){
    String msg = "0 ";
    msg += timePacket[k];
    msg += " ";
    msg += bufferPacket[k];
    msg += " 70000";
  }
}

void print_status()
{
  Serial.print("Status message:\n");
  switch(Wire.status())
  {
  case I2C_WAITING:
      Serial.print("Waiting for action/cmd.\n");
      break;
  case I2C_SENDING: 
      Serial.print("Sending input.\n");
      break;
  case I2C_SEND_ADDR: 
      Serial.print("Sending address.\n");
      break;
  case I2C_RECEIVING: 
      Serial.print("Receiving data.\n");
      break;
  case I2C_TIMEOUT: 
      Serial.print("Timeout has occurred.\n");
      break;
  case I2C_ADDR_NAK: 
      Serial.print("No acknowledge bit for address action.\n");
      break;
  case I2C_DATA_NAK: 
      Serial.print("No acknowledge bit for data action.\n");
      break;
  case I2C_ARB_LOST: 
      Serial.print("ARB_LOST status.\n");
      break;
  case I2C_SLAVE_TX: 
      Serial.print("Transmitting as a slave.\n");
      break;
  case I2C_SLAVE_RX: 
      Serial.print("Receiving as a slave.\n");
      break;
  default:
      Serial.print("Unknown error.\n");
      break;
  }
  Serial.print("\n");
}

void print_errors()
{
  Serial.print("Error message:\n");
  switch(Wire.getError())
  {
  case 0:
      Serial.print("No errors found.\n");
      break;
  case 1: 
      Serial.print("Data too long.\n");
      break;
  case 2:
      Serial.print("Received no acknowledge bit from address action.\n");
      break;
  case 3:
      Serial.print("Received no acknowledge bit from data action.\n");
      break;
  case 4:
      Serial.print("Other error has occurred.\n");
      break;
  default:
      Serial.print("Unknown error.\n");
      break;
  }
  Serial.print("\n");
}

void calibration(){
    Serial.print("Calibration routine begins. \n");
    unsigned long duration = 6000; // 6 sec
    unsigned long startTime = millis();
    while (millis() < (startTime + duration)) {
    int sensorValue = analogRead(sensorPin);

    // record the maximum sensor value
    if (sensorValue > sensorMax) {
      sensorMax = sensorValue;
    }

    // record the minimum sensor value
    if (sensorValue < sensorMin) {
      sensorMin = sensorValue;
    }
  }
  Serial.printf("Max ADC level is %d \n",sensorMax);
  Serial.printf("Min ADC level is %d \n",sensorMin);
}

void readFromPGA(){
  Wire.write(currAddr);
  // Print message
  Serial.print("Reading from slave. \n");
  Serial.printf("Read address is 0x%X.\n", currAddr);
  // Read from Slave (is this the right format?)
  //Wire.requestFrom(target, 2, false); // Restart condition w/ request
  Wire.requestFrom(target, 2, I2C_NOSTOP);
  Serial.printf("There are %d bytes available from PGA. \n", Wire.available());
  while(Wire.available()){
    //Wire.read()
    byte2 = Wire.readByte();
    byte1 = Wire.readByte();
  }
  // Temperature reading
  if(isTempADC){
    float temp = tempConvert(byte1,byte2);
    Serial.printf("The current temperature is %f degrees Celsius. \n", temp);
  // Error code reading
  } else {
    uint16_t errorCode = concatBytes(byte1,byte2);
    Serial.printf("The fault code is %d. \n", errorCode);
  }
}

void writeToPGA(){
    // Print message to register to modify amp behavior
  Serial.print("Sending to slave.\n");
  Wire.write(currAddr);
  Serial.printf("Write address is 0x%X.\n", currAddr);
  // Has to be two addresses
  // Address 1-8, then address 9-16
  rawTwoByte = writeSelect(currAddr);
  splitTwoByte(writeArray,rawTwoByte);
  Wire.write(writeArray,2); // transmits two bytes from an array
  Serial.printf("Raw write data is 0x%X.\n", rawTwoByte);
  Serial.printf("First byte sent is 0x%X.\n", writeArray[1]);
  Serial.printf("Second byte sent is 0x%X.\n", writeArray[0]);
}

// Converts ADC temperature readings into usable float data
float tempConvert(uint8_t byte1, uint8_t byte2){
  // Concat the two bytes
  uint16_t count = concatBytes(byte1,byte2);
  // Multiply the 2-byte by the multiplier to yield Celsius reading
  float tempC = 0.0625 * count;
  return tempC;
}

uint16_t concatBytes(uint8_t byte1, uint8_t byte2){
  uint16_t twoByte;
  twoByte = byte1;
  twoByte = twoByte<<8;
  twoByte |= byte2;
  Serial.printf("The raw read data is 0x%X. \n", twoByte);
  return twoByte;
}

void splitTwoByte(uint8_t *pWrite, uint16_t twoByte){
  uint8_t botByte = (uint8_t) (twoByte & 0xFF);
  uint8_t topByte = (uint8_t) ((twoByte >> 8) & 0xFF);
  pWrite[1] = topByte;
  pWrite[0] = botByte;
}

// Selects an internal register to write to
uint16_t writeSelect(uint8_t currAddr){
  uint16_t writeMsg;
  switch(currAddr) {
    case 0x01:
    writeMsg = fineOffsetAdjust();
    break; 
    case 0x02:
    writeMsg = fineGainAdjust();
    break;
    case 0x03:
    writeMsg = refCtrlLinear();
    break;
    case 0x04:
    writeMsg = PGAOffsetGain();
    break;
    case 0x05:
    writeMsg = PGAConfigLimit();
    break;    
    case 0x06:
    writeMsg = TempADCCtrl();
    break;
    case 0x07:
    writeMsg = OutEnableCounterCtrl();
    break;
    default:
    writeMsg = fineOffsetAdjust();
  }
  return writeMsg;
}

uint16_t fineOffsetAdjust(){
  // Assume 5 volt reference
  float V_ref = 5.0;
  float V_zero = 1.25;
  float ZD_float = V_zero * 65536 / V_ref;
  uint16_t ZD = (uint16_t) ZD_float;
  return ZD;
}

uint16_t fineGainAdjust(){
  //float targetG = 0.5;
  float targetG = 0.85;
  float GD_float = (targetG - 0.33333333) * 1.5 * 65536;
  uint16_t GD = (uint16_t) GD_float;
  return GD;
}

uint16_t refCtrlLinear(){
  uint16_t result;

  uint8_t RFB = 0x00;

  // Linear Adjust, V_exc gain select
  // 0x00 : -0.166 V_fb to +0.166 V_fb lineariz. DAC; 0.83 V_ref of V_exc gain
  // 0x01 : -0.124 V_fb to +0.124 V_fb lineariz. DAC; 0.52 V_ref of V_exc gain
  uint8_t EXS = 0x00;

  // Enable V_exc
  uint8_t EXEN = 0x00;

  // Select internal V_ref (0-4.096V, 1-2.5V)
  uint8_t RS = 0x00;

  // Enable internal V_ref
  uint8_t REN = 0x00;

  float FSR;
  if(EXS == 0x01){
    FSR = 0.166;
  }else{
    FSR = 0.124;
  }
  
  float desiredFBRatio = 0.04183; // ratio
  int8_t LD = (int8_t) (desiredFBRatio / (FSR / 127));

  // Starts with 4 bits
  result = RFB; result = result<<1;
  result |= EXS; result = result<<1;
  result |= EXEN; result = result<<1;
  result |= RS; result = result<<1;
  result |= REN; result = result<<8;
  result |= LD;
  
  return result;
}

uint16_t PGAOffsetGain(){  
  uint16_t result;
  
  uint8_t OWD = 0x01; // maybe not necessary for two-wire?

  // Output Amp gain table
  // 0x00 - 2
  // 0x01 - 2.4
  // 0x02 - 3
  // 0x03 - 3.6
  // 0x04 - 4.5
  // 0x05 - 6
  // 0x06 - 9
  // 0x07 - Disable internal feedback
  uint8_t GO = 0x00;

  // 0x00 - V_in1 = V_inp; V_in2 = V_inn;
  // 0x01 - V_in1 = V_inn; V_in1 = V_inp;
  uint8_t GI3 = 0x00;

  // Frontend gain table
  // 0x00 - 4
  // 0x01 - 8
  // 0x02 - 16
  // 0x03 - 23.27
  // 0x04 - 32
  // 0x05 - 42.67
  // 0x06 - 64
  // 0x07 - 128
  uint8_t GI = 0x00;
  
  uint8_t RFB = 0x00;

  // Assume 5 volt reference
  float V_ref = 5.0;
  float desiredOffset = 17; // mV
  // make signed (!!!!!)
  uint8_t OS = (uint8_t) (desiredOffset / (V_ref * 0.85));
  
 
  // Starts with 1 bits
  result = OWD; result = result<<3;
  result |= GO; result = result<<1;
  result |= GI3; result = result<<3;
  result |= GI; result = result<<3;
  result |= RFB; result = result<<5;
  result |= OS;
  
  return result;
}

uint16_t PGAConfigLimit(){
  uint16_t result;
  
  uint8_t RFB = 0x00;

  // Table 6-11 in User manual
  uint8_t CLK_CFG = 0x00;

  // Enable ext fault comp group
  uint8_t EXTEN = 0x00;

  // Enable int fault comp group
  uint8_t INTEN = 0x00;

  // V_out polarity (1-high, 0-low) when ext fault detected
  uint8_t EXTPOL = 0x00;

  // V_out polarity (1-high, 0-low) when int fault detected
  uint8_t INTPOL = 0x00;

  // Over/under-scale Limit Enable
  uint8_t OUEN = 0x00;
  
  // 0x00 - 0.9708 V_ref
  // 0x01 - 0.9610 V_ref
  // 0x02 - 0.9394 V_ref
  // 0x03 - 0.9160 V_ref
  // 0x04 - 0.9102 V_ref
  // 0x05 - 0.7324 V_ref
  // 0x06 - 0.5528 V_ref
  // 0x07 - Reserved
  uint8_t HL = 0x00;

  // 0x00 - 0.02540 V_ref
  // 0x01 - 0.02930 V_ref
  // 0x02 - 0.03516 V_ref
  // 0x03 - 0.03906 V_ref
  // 0x04 - 0.04492 V_ref
  // 0x05 - 0.05078 V_ref
  // 0x06 - 0.05468 V_ref
  // 0x07 - 0.06054 V_ref
  uint8_t LL = 0x00;

  // Starts with 2 bits
  result = RFB; result = result<<2;
  result |= CLK_CFG; result = result<<1;
  result |= EXTEN; result = result<<1;
  result |= INTEN; result = result<<1;
  result |= EXTPOL; result = result<<1;
  result |= INTPOL; result = result<<1;
  result |= RFB; result = result<<1;
  result |= OUEN; result = result<<3;
  result |= HL; result = result<<3;
  result |= LL;

  return result;
}

uint16_t TempADCCtrl(){
  uint16_t result;
  
  uint8_t RFB = 0x00;
  
  // Enable twice conversion speed for temp ADC
  uint8_t ADC2X = 0x00;
  
  // Start or restart the ADC
  uint8_t ADCS = 0x01;
  
  // Enable 7microA TEMP_in current source (I_temp)
  uint8_t ISEN = 0x00;
  
  // Enable continuous conversion
  uint8_t CEN = 0x00;
  
  // Enable Internal temp mode
  // Set ADC2X to 0; AREN to 0; RV, M, G to 00;
  uint8_t TEN = 0x00;
  
  // Enable internal 2.048 V reference, ignores RV
  uint8_t AREN = 0x00;
  
  // External temp ref
  // 0x00 - V_ref
  // 0x01 - V_exc
  // 0x02 - V_SA
  // 0x03 - Reserved
  uint8_t RV = 0x00;
  
  // Temp ADC Mux Select
  // 0x00 - TEMP_in (+), GND_a (-)
  // 0x01 - V_exc (+), TEMP_in (-)
  // 0x02 - V_out (+), GND_a (-)
  // 0x03 - V_ref (+), TEMP_in (-) 
  uint8_t MUX = 0x00;
  
  // Temp ADC PGA Gain
  // 0x00 - 1
  // 0x01 - 2
  // 0x02 - 4
  // 0x03 - 8 
  uint8_t GAIN = 0x00;

  // Refer to Table 6-17 in PGA309 User Manual (it's too long to comment O_o)
  uint8_t RES = 0x00;

  // Starts with 2 bits
  result = RFB; result = result<<1;
  result |= ADC2X; result = result<<1;
  result |= ADCS; result = result<<1;
  result |= ISEN; result = result<<1;
  result |= CEN; result = result<<1;
  result |= TEN; result = result<<1;
  result |= AREN; result = result<<2;
  result |= RV; result = result<<2;
  result |= MUX; result = result<<2;
  result |= GAIN; result = result<<2;
  result |= RES;
  
  return result;
}

uint16_t OutEnableCounterCtrl(){
  uint16_t result;
  
  uint8_t RFB = 0x00;
  
  // Temp ADC Delay (OEN decimal value * 14) ms
  int tempADCDelay = 98; // ms
  uint8_t DLY = tempADCDelay / 14;
  
  // V_out enable timeout
  int desiredTimeout = 2240; // ms
  uint8_t OEN = desiredTimeout / 14;

  // Starts with 4 bits
  result = RFB; result = result<<4;
  result |= DLY; result = result<<8;
  result |= OEN;  
  
  return result;
}

